// dans Makefile de drivers/char : obj-m    += test_jm

// TODO : lancer la conversion par un ioctl qui passe le nbre de points en arg, et 
//        ne faire que recuperer les points dans le read
/*
 * 12:42 < sim> jmfriedt: je m'etais inspire du chapitre que tu as lu pour le driver qadc
 * 12:42 < sim> jmfriedt: mais je crois que j'ai pas fait le mieux possible pour le buffer circulaire
 * 12:43 < sim> jmfriedt: je me demande si il n'y a pas des outils tout fait dans le noyau
 * 12:46 < sim> jmfriedt: http://lwn.net/Articles/107314/
 * 12:46 < sim> jmfriedt: regarde "Simple circular buffers"
 * 12:47 < sim> jmfriedt: fait par Stelian Pop :)
 * 12:47 < sim> kfifo_init, kfifo_get, kfifo_put
 * 12:47 < sim> jmfriedt: c'est exactement ce que tu veux
 * 12:49 < sim> Even a data structure as simple as a circular buffer can be tricky to implement correctly; it makes no 
 *              sense for every developer to go through that process each time a new one is needed. With a single, 
                well-reviewed implementation, the chances of it being truly correct are much better.
 * 12:49 < sim> jmfriedt: je suis d'accord a 200%
 * 12:50 < sim> jmfriedt: include/linux/kfifo.h
 * 13:01 < sim> jmfriedt: linux-2.6.git/net/dccp/probe.c
 * 13:02 < sim> jmfriedt: lis la fonction dccpprobe_read()
 * 13:02 < sim> jmfriedt: c'est l'implementation parfaite pour ton convertisseur
 * 13:02 < sim> jmfriedt: il te reste le gestionnaire d'interruption a faire sur le meme model
 */
// http://www.captain.at/howto-linux-device-driver-template-skeleton.php
// Linux Device Driver Template/Skeleton
// Kernel Module

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#define SKELETON_MAJOR 32
#define SKELETON_NAME "skeleton"
#define CART_INTERRUPT 13 // jmfriedt
#define CASE1 1
#define CASE2 2

#undef jmf_debug

static char string [256];
static int interruptcount = 0;
static int count=0;

DECLARE_WAIT_QUEUE_HEAD(skeleton_wait);

// open function - called when the "file" /dev/skeleton is opened in userspace
static int skeleton_open (struct inode *inode, struct file *file) {
	printk("skeleton_open\n");
	return 0;
}

// ioctl - I/O control
static int skeleton_ioctl(struct inode *inode, struct file *file,
		unsigned int cmd, unsigned long arg) {
	int retval = 0;
	int len, oldcount;
	wait_queue_t wait;

	switch ( cmd ) {
		case CASE1:/* for writing data to arg */
			if (copy_from_user(&count, (int *)arg, sizeof(int)))
			return -EFAULT;
	                printk("ioctl: count=%d\n",count);

// race condition decrite a
// http://www.xml.com/ldd/chapter/book/ch09.html#t8

	                init_waitqueue_entry(&wait, current);
                        interruptcount=0,oldcount=0;

	                // check if we have data - if not, sleep & wake up in interrupt_handler
	                while (interruptcount<count)
                          {*(unsigned short*)(0x8000000)=(unsigned short)0;
//   	     interruptible_sleep_on(&skeleton_wait);             <-- race !
	                   add_wait_queue(&skeleton_wait, &wait);
                           while (1) {
                             set_current_state(TASK_INTERRUPTIBLE);
                             if (interruptcount != oldcount ) 
                                 break;
                             schedule();
                           }
                           set_current_state(TASK_RUNNING);
                           remove_wait_queue(&skeleton_wait, &wait);
                           oldcount=interruptcount;

	                   string[interruptcount-1]=*(char*)(0x8000000)&0xff;
	                  }
	                  len  = interruptcount;
                          printk(">>> Slot2 INT HANDLED: interruptcount=%d\n", interruptcount);
                          interruptcount=0;

			break;
/*		case CASE2:// for reading data from arg 
			if (copy_to_user((int *)arg, &data, sizeof(int)))
			return -EFAULT;
			break;
*/		default:
			retval = -EINVAL;
	}
	return retval;
}

// close function - called when the "file" /dev/skeleton is closed in userspace  
static int skeleton_release (struct inode *inode, struct file *file) {
	printk("skeleton_release\n");
	return 0;
}

// read function called when from /dev/skeleton is read
static ssize_t skeleton_read (struct file *file, char *buf,
		size_t readcount, loff_t *ppos) {
	int err=0;
        if (count>readcount) 
	   err = copy_to_user(buf,string,readcount);
	if (err != 0) return -EFAULT;
	count=0;
	return readcount;
}

// interrupt handler
static irqreturn_t interrupt_handler(int irq, void *dummy)
{
	interruptcount++;
#ifdef jmf_debug
        printk(">>> Slot2 INT HANDLED: interruptcount=%d\n", interruptcount);
#endif
	// wake up (unblock) for reading data from userspace
	wake_up_interruptible(&skeleton_wait);
	
	return IRQ_HANDLED;
}

// define which file operations are supported
struct file_operations skeleton_fops = {
	.owner	=	THIS_MODULE,
	.llseek	=	NULL,
	.read	=	skeleton_read,
	.write	=	NULL,
	.readdir=	NULL,
	.poll	=	NULL,
	.ioctl	=	skeleton_ioctl,
	.mmap	=	NULL,
	.open	=	skeleton_open,
	.flush	=	NULL,
	.release=	skeleton_release,
	.fsync	=	NULL,
	.fasync	=	NULL,
	.lock	=	NULL,
	.readv	=	NULL,
	.writev	=	NULL,
};

// initialize module (and interrupt)
static int __init skeleton_init_module (void) {
	int i;
	int ret;
	printk("initializing module\n");
	
	i = register_chrdev (SKELETON_MAJOR, SKELETON_NAME, &skeleton_fops);
	if (i != 0) return - EIO;
	
	// INIT INTERRUPT
	ret=request_irq(CART_INTERRUPT, interrupt_handler, SA_INTERRUPT, "SLOT2cart", NULL);
	enable_irq(13); printk("interrupt enabled\n");
	return 0;
}

// close and cleanup module
static void __exit skeleton_cleanup_module (void) {
	printk("cleaning up module\n");
	unregister_chrdev (SKELETON_MAJOR, SKELETON_NAME);
	disable_irq(13);
	free_irq(13, interrupt_handler);
}

module_init(skeleton_init_module);
module_exit(skeleton_cleanup_module);
MODULE_AUTHOR("www.captain.at");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Linux Device Driver Template/Skeleton with Interrupt Handler");

